/*
 * Test dialog with ListView.
 * Monitors predefined processes for a possibility to allocate memory with VirtuallAllocEx.
 * 64bit version allocate all possible memory in 32bit address space.
 *
 * AZ, 2020
 *
 * This file is covered by the GNU General Public License v2.
 */
#define UNICODE
#define _UNICODE
#include <windows.h>
#include <olectl.h>
#include <shlwapi.h>
#include <stdio.h>
#include <stdarg.h>
#include <tchar.h>
#include <commctrl.h>
#include <tlhelp32.h>

#if defined(__x86_64__)
#define ATEST64
#else
#define ATEST32
#endif

#define IDC_LISTBOX 0x1000
#define IDC_REAPER  0x1001
#ifdef ATEST32
#define IDC_ATEST64 0x1002
#endif

#ifdef ATEST64
#define ADDR_64 ((void *)(0x100000000LL))
#endif

static void CreateChildWindows(HINSTANCE hInstance, HWND hTop){
    static TCHAR *szItems[] = { TEXT("Item 1"), TEXT("Item 2"), TEXT("Item 3") };
    HWND hList;
    RECT rTop;
    LVCOLUMN lvc;
    LVITEM   lvi = {0};
    unsigned n;


    GetClientRect(hTop, &rTop);

    hList = CreateWindow(WC_LISTVIEW, TEXT("Test listview"),
			 WS_VISIBLE | WS_CHILD | WS_BORDER | LVS_REPORT | LVS_NOCOLUMNHEADER | LVS_SINGLESEL | LVS_SHOWSELALWAYS | WS_GROUP | WS_TABSTOP,
                         10, 10, rTop.right - 20, rTop.bottom - 20 - 40, hTop, (HMENU)IDC_LISTBOX, hInstance, NULL);
    ListView_SetExtendedListViewStyle(hList, LVS_EX_FULLROWSELECT);

    lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
    lvc.fmt = LVCFMT_LEFT;
    lvc.cx = rTop.right - 30;
    lvc.pszText = TEXT("Column");
    lvc.iSubItem = 0;
    ListView_InsertColumn(hList, 0, &lvc);

    lvi.mask = LVIF_TEXT;
    for(n = 0; n < sizeof(szItems)/sizeof(*szItems); ++n){
      lvi.iItem = n;
      lvi.pszText = szItems[n];
      ListView_InsertItem(hList, &lvi);
    }

    ListView_SetItemState(hList, 0,  LVIS_FOCUSED | LVIS_SELECTED, LVIS_FOCUSED | LVIS_SELECTED);

    CreateWindow(TEXT("STATIC"), TEXT(""), WS_VISIBLE | WS_CHILD | SS_CENTER | WS_GROUP | WS_TABSTOP,
		 10, rTop.bottom - 20 - 40 + 10, rTop.right - 20, 20, hTop, (HMENU)IDC_REAPER, hInstance, NULL);
#ifdef ATEST32
    CreateWindow(TEXT("STATIC"), TEXT(""), WS_VISIBLE | WS_CHILD | SS_CENTER | WS_GROUP | WS_TABSTOP,
		 10, rTop.bottom - 20 - 40 + 30, rTop.right - 20, 20, hTop, (HMENU)IDC_ATEST64, hInstance, NULL);
#endif
}

static HANDLE FindApp(TCHAR *szAppName){
  PROCESSENTRY32 pe;
  HANDLE ts = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  HANDLE hProcess = NULL;

  pe.dwSize = sizeof(PROCESSENTRY32);

  if(Process32First(ts, &pe)){
    do {
      if(!_tcsicmp(pe.szExeFile, szAppName)){
	hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, pe.th32ProcessID);
	break;
      }
    } while(Process32Next(ts, &pe));
  }
  CloseHandle(ts);
  return hProcess;
}

static void UpdateAppInfo(HWND hTop, int dCtlId, TCHAR *szAppName){
  HANDLE hProcess;
  void *vmem;
  TCHAR szText[1024], szOldText[1024];
  if((hProcess = FindApp(szAppName))){
    vmem = VirtualAllocEx(hProcess, NULL, 4, MEM_COMMIT, PAGE_READWRITE);
    if(vmem){
      VirtualFreeEx(hProcess, vmem, 0, MEM_RELEASE);
    }
    CloseHandle(hProcess);
    if(!vmem)
      _sntprintf(szText, 1024, TEXT("Can not allocate memory in %s: %u"), szAppName, GetLastError());
    else {
#ifdef ATEST64
      if(vmem >= ADDR_64)
	_sntprintf(szText, 1024, TEXT("%s memory allocated at high address %p"), szAppName, vmem);
      else
#endif
	_sntprintf(szText, 1024, TEXT("%s memory allocated at %p"), szAppName, vmem);
    }
  } else
    _sntprintf(szText, 1024, TEXT("%s is not running"), szAppName);
  GetWindowText(GetDlgItem(hTop, dCtlId), szOldText, 1024);
  if(_tcscmp(szText, szOldText))
    SetWindowText(GetDlgItem(hTop, dCtlId), szText);
}

#ifdef ATEST64

/*
 * Reserve or release all 32bit addresses
 */

typedef struct _VMemBlock {
  struct _VMemBlock *next;
  void *p;
  size_t size;
} VMemBlock;

static void ReserveLowerMemory(int reserve){
  static VMemBlock *blocks = NULL;
  HANDLE hProcess = GetCurrentProcess();

  VMemBlock *pblk;
  size_t step = 1024*1024*256;
  //size_t step = 8;
  void *vmem;
  if(!reserve){
    while((pblk = blocks)){
      blocks = pblk->next;
      VirtualFreeEx(hProcess, pblk->p, 0, MEM_RELEASE);
      free(pblk);
    }
    return;
  }
  vmem = VirtualAllocEx(GetCurrentProcess(), NULL, 4, MEM_COMMIT, PAGE_READWRITE);
  if(!vmem)
    return; // can not allocate at all, that is an error...
  VirtualFreeEx(hProcess, vmem, 0, MEM_RELEASE);
  if(vmem >= ADDR_64)
    return; // can allocate, but address is in the upper addresses space
  while(1){
    vmem = VirtualAllocEx(GetCurrentProcess(), NULL, step, MEM_COMMIT, PAGE_READWRITE);
    if(!vmem)
      return; // can not allocate at all, that is an error...
    if(vmem >= ADDR_64){
      VirtualFreeEx(hProcess, vmem, 0, MEM_RELEASE);
      step >>= 2;
      if(step < 4)
	return; // can not allocate a single integer in the lower address space
      continue; // try with lower size
    }

    pblk = (VMemBlock *)malloc(sizeof(VMemBlock));
    if(!pblk){
      VirtualFreeEx(hProcess, vmem, 0, MEM_RELEASE);
      return; // that is a bug...
    }
    pblk->next = blocks;
    pblk->p = vmem;
    pblk->size = step;
    blocks = pblk;
    printf("VMEM: %p [%u]\n", vmem, (unsigned)step);
  }
}

#endif


#define IDC_TIMER 0x1

/*
 * Periodically check that "lower memory" addresses are reserved when asked
 */
static void OnTimer(HWND hTop){
  UpdateAppInfo(hTop, IDC_REAPER, TEXT("REAPER.EXE"));
#ifdef ATEST64
  ReserveLowerMemory(1);
#else
  UpdateAppInfo(hTop, IDC_ATEST64, TEXT("ATEST.EXE"));
#endif
  SetTimer(hTop, IDC_TIMER, 2000, NULL);
}

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

static HWND hDlgCurrent = NULL;


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   PSTR szCmdLine, int iCmdShow){
  static TCHAR const szAppName[] = TEXT("ATest");
  HWND hWnd;
  MSG msg;
  WNDCLASSEX wcex = { 0 };
  int ret;

#ifdef ATEST64
  // so we can "printf"
  FILE* fp;
  AllocConsole();
  freopen_s(&fp, "CONIN$", "r", stdin);
  freopen_s(&fp, "CONOUT$", "w", stdout);
  freopen_s(&fp, "CONOUT$", "w", stderr);
#endif

  // for WIC
  // CoInitializeEx(NULL, COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE);

  wcex.cbSize = sizeof(WNDCLASSEX);
  wcex.style = CS_HREDRAW | CS_VREDRAW;
  wcex.lpfnWndProc   = &WndProc;
  wcex.cbClsExtra = 0;
  wcex.cbWndExtra = 0;
  wcex.hInstance = hInstance;
  wcex.hIcon = LoadIcon(NULL, IDI_APPLICATION);
  wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
  //wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
  wcex.hbrBackground = (HBRUSH)(COLOR_3DFACE+1);
  wcex.lpszMenuName = NULL;
  wcex.lpszClassName = szAppName;
  wcex.hIconSm = wcex.hIcon;

  if( !RegisterClassEx(&wcex) ){
    MessageBox(NULL, TEXT("RegisterClassEx had problems"),
               szAppName, MB_OK | MB_ICONERROR);
    return -1;
  }
  InitCommonControls();
  hWnd = CreateWindowEx(WS_EX_OVERLAPPEDWINDOW, szAppName,
#ifdef ATEST64
                        TEXT("Accessibility test 64bit"),
#else
                        TEXT("Accessibility test 32bit"),
#endif
                        WS_OVERLAPPEDWINDOW, CW_USEDEFAULT,CW_USEDEFAULT,
                        500, 300, NULL,NULL,hInstance,NULL);
  if(!hWnd){
    MessageBox(NULL, TEXT("CreateWindowEx had problems"),
               szAppName, MB_OK | MB_ICONERROR);
    return -2;
  }
  CreateChildWindows(hInstance,hWnd);
  ShowWindow(hWnd, iCmdShow);
  UpdateWindow(hWnd);
  OnTimer(hWnd);
  while( (ret=GetMessage( &msg, NULL, 0, 0 )) > 0 ){
    if (NULL == hDlgCurrent || !IsDialogMessage(hDlgCurrent, &msg)){
      TranslateMessage(&msg);
      DispatchMessage( &msg );
    }
  }
  if(ret < 0)
    return ret;

  UnregisterClass(szAppName, hInstance);
  return (int)msg.wParam;
}

static void OnPaint(HWND hWnd, HDC hDC, RECT *rcUpdate){
  //TextOut(hDC, 0, 0, TEXT("Knob Test ;)"), 12);
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message,
                         WPARAM wParam, LPARAM lParam){
  PAINTSTRUCT ps;

  switch(message){
  case WM_ACTIVATE:
    if(!LOWORD(wParam)) // becoming inactive
      hDlgCurrent = NULL;
    else { // becoming active
      hDlgCurrent = hWnd;
      SetFocus(GetDlgItem(hWnd, IDC_LISTBOX));
    }
    return FALSE;

  case WM_TIMER:
    OnTimer(hWnd);
    break;

  case WM_PRINTCLIENT:
    OnPaint(hWnd, (HDC)wParam, NULL);
    break;

  case WM_PAINT:
    BeginPaint(hWnd, &ps);
    OnPaint(hWnd, ps.hdc, &ps.rcPaint);
    EndPaint(hWnd, &ps);
    return 0;

  case WM_CLOSE:
    DestroyWindow(hWnd);
    break;

  case WM_DESTROY:
    PostQuitMessage(0);
    return 0;
  }

  return DefWindowProc(hWnd, message, wParam, lParam);
}



